import pytest

from kestrel.session import Session


def test_describe():
    with Session() as s:
        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": 123}
             , {"type": "process", "name": "explorer.exe", "pid": 99}
             , {"type": "process", "name": "explorer.exe", "pid": 200}
             ]
"""
        s.execute(stmt)
        out = s.execute("DESCRIBE newvar.name")
        stats = out[0].to_dict()['data']
        assert stats['count'] == 3
        assert stats['unique'] == 2
        assert stats['top'] == "explorer.exe"
        assert stats['freq'] == 2

        out = s.execute("DESCRIBE newvar.pid")
        stats = out[0].to_dict()['data']
        assert stats['count'] == 3
        assert stats['mean'] == (123 + 99 + 200)/3
        assert stats['min'] == 99
        assert stats['max'] == 200



import time
import logging
from multiprocessing import Process, Queue, current_process
from typing import Optional
from typeguard import typechecked

from stix_shifter.stix_transmission import stix_transmission
from kestrel_datasource_stixshifter.worker import STOP_SIGN
from kestrel_datasource_stixshifter.worker.utils import TransmissionResult, WorkerLog


@typechecked
class TransmitterPool(Process):
    def __init__(
        self,
        connector_name: str,
        connection_dict: dict,
        configuration_dict: dict,
        retrieval_batch_size: int,
        number_of_translators: int,
        cool_down_after_transmission: int,
        queries: list,
        output_queue: Queue,
        limit: Optional[int],
    ):
        super().__init__()

        self.connector_name = connector_name
        self.connection_dict = connection_dict
        self.configuration_dict = configuration_dict
        self.retrieval_batch_size = retrieval_batch_size
        self.number_of_translators = number_of_translators
        self.cool_down_after_transmission = cool_down_after_transmission
        self.queries = queries
        self.queue = output_queue
        self.limit = limit

    def run(self):
        transmitters = [
            Transmitter(
                self.connector_name,
                self.connection_dict,
                self.configuration_dict,
                self.retrieval_batch_size,
                self.cool_down_after_transmission,
                query,
                self.queue,
                self.limit,
            )
            for query in self.queries
        ]
        for transmitter in transmitters:
            transmitter.start()
        for transmitter in transmitters:
            transmitter.join()
        for _ in range(self.number_of_translators):
            self.queue.put(STOP_SIGN)


class Transmitter(Process):
    def __init__(
        self,
        connector_name: str,
        connection_dict: dict,
        configuration_dict: dict,
        retrieval_batch_size: int,
        cool_down_after_transmission: int,
        query: str,
        output_queue: Queue,
        limit: Optional[int],
    ):
        super().__init__()

        self.connector_name = connector_name
        self.connection_dict = connection_dict
        self.configuration_dict = configuration_dict
        self.retrieval_batch_size = retrieval_batch_size
        self.cool_down_after_transmission = cool_down_after_transmission
        self.query = query
        self.queue = output_queue
        self.limit = limit

    def run(self):
        self.worker_name = current_process().name
        self.transmission = stix_transmission.StixTransmission(
            self.connector_name,
            self.connection_dict,
            self.configuration_dict,
        )
        search_meta_result = self.transmission.query(self.query)

        if search_meta_result["success"]:
            self.search_id = search_meta_result["search_id"]
            if self.wait_datasource_search():
                # no error so far
                self.retrieve_data()
        else:
            err_msg = (
                search_meta_result["error"]
                if "error" in search_meta_result
                else "details not avaliable"
            )
            packet = TransmissionResult(
                self.worker_name,
                False,
                None,
                None,
                WorkerLog(
                    logging.ERROR,
                    f"STIX-shifter transmission.query() failed: {err_msg}",
                ),
            )
            self.queue.put(packet)

    def wait_datasource_search(self):
        # kestrel init status: "KINIT"
        status = {"success": True, "progress": 0, "status": "KINIT"}

        while (
            status["success"]
            and status["progress"] < 100
            and status["status"] in ("KINIT", "RUNNING")
        ):
            if status["status"] == "KINIT":
                time.sleep(self.cool_down_after_transmission)
            elif status["status"] == "RUNNING":
                time.sleep(max(1, self.cool_down_after_transmission))
            status = self.transmission.status(self.search_id)
            if not status["success"]:
                err_msg = (
                    status["error"] if "error" in status else "details not avaliable"
                )
                packet = TransmissionResult(
                    self.worker_name,
                    False,
                    None,
                    None,
                    WorkerLog(
                        logging.ERROR,
                        f"STIX-shifter transmission.status() failed: {err_msg}",
                    ),
                )
                self.queue.put(packet)
                return False
        return True

    def retrieve_data(self):
        result_retrieval_offset = 0
        has_remaining_results = True
        metadata = None
        is_retry_cycle = False
        batch_size = self.retrieval_batch_size
        if self.limit and self.limit < self.retrieval_batch_size:
            batch_size = self.limit

        while has_remaining_results:
            packet = None
            time.sleep(self.cool_down_after_transmission)
            result_batch = self.transmission.results(
                self.search_id,
                result_retrieval_offset,
                batch_size,
                metadata,
            )

            if result_batch["success"]:
                if result_batch["data"]:
                    packet = TransmissionResult(
                        self.worker_name,
                        True,
                        result_batch["data"],
                        result_retrieval_offset,
                        None,
                    )

                    # prepare for next round retrieval
                    result_retrieval_offset += len(result_batch["data"])
                    if "metadata" in result_batch:
                        metadata = result_batch["metadata"]

                    if self.limit:
                        if result_retrieval_offset >= self.limit:
                            has_remaining_results = False
                        else:
                            batch_size = self.limit - result_retrieval_offset
                            if batch_size > self.retrieval_batch_size:
                                batch_size = self.retrieval_batch_size
                else:
                    has_remaining_results = False

                is_retry_cycle = False

            else:
                err_msg = (
                    result_batch["error"]
                    if "error" in result_batch
                    else "details not avaliable"
                )

                if (
                    err_msg.startswith(
                        f"{self.connector_name} connector error => server timeout_error"
                    )
                    and not is_retry_cycle
                ):
                    # mitigate https://github.com/opencybersecurityalliance/stix-shifter/issues/1493
                    # only give it one retry to mitigate high CPU occupation
                    # otherwise, it could be a real server connection issue
                    # /stix_shifter_utils/stix_transmission/utils/RestApiClientAsync.py
                    packet = TransmissionResult(
                        self.worker_name,
                        False,
                        None,
                        None,
                        WorkerLog(
                            logging.INFO,
                            "Busy CPU; hit stix-shifter aiohttp connection timeout; retry.",
                        ),
                    )
                    is_retry_cycle = True

                else:
                    packet = TransmissionResult(
                        self.worker_name,
                        False,
                        None,
                        None,
                        WorkerLog(
                            logging.ERROR,
                            f"STIX-shifter transmission.result() failed: {err_msg}",
                        ),
                    )
                    has_remaining_results = False

            if packet:
                self.queue.put(packet)



import os
import pytest
import pathlib
from kestrel.codegen.relations import all_relations
from kestrel.session import Session
from kestrel.syntax.utils import (
    LITERALS,
    AGG_FUNCS,
    TRANSFORMS,
    EXPRESSION_OPTIONS,
)


KNOWN_ETYPES = {
    "artifact",
    "autonomous-system",
    "directory",
    "domain-name",
    "email-addr",
    "email-message",
    "file",
    "ipv4-addr",
    "ipv6-addr",
    "mac-addr",
    "mutex",
    "network-traffic",
    "process",
    "software",
    "url",
    "user-account",
    "windows-registry-key",
    "x-ibm-finding",
    "x-oca-asset",
    "x-oca-event",
}


@pytest.fixture
def datasource_env_setup(tmp_path):

    profiles = f"""profiles:
    thost101:
        connector: elastic_ecs
        connection:
            host: elastic.securitylog.company.com
            port: 9200
            selfSignedCert: false # this means do NOT check cert
            indices: asdfqwer
        config:
            auth:
                id: VuaCfGcBCdbkQm-e5aOx
                api_key: ui2lp2axTNmsyakw9tvNnw
    thost102:
        connector: qradar
        connection:
            host: qradar.securitylog.company.com
            port: 443
        config:
            auth:
                SEC: 123e4567-e89b-12d3-a456-426614174000
    thost103:
        connector: cbcloud
        connection:
            host: cbcloud.securitylog.company.com
            port: 443
        config:
            auth:
                org-key: D5DQRHQP
                token: HT8EMI32DSIMAQ7DJM
    """

    profile_file = tmp_path / "stixshifter.yaml"
    with open(profile_file, "w") as pf:
        pf.write(profiles)

    os.environ["KESTREL_STIXSHIFTER_CONFIG"] = str(
        profile_file.expanduser().resolve()
    )

    # https://docs.pytest.org/en/latest/how-to/fixtures.html#teardown-cleanup-aka-fixture-finalization
    yield None
    del os.environ["KESTREL_STIXSHIFTER_CONFIG"]


@pytest.fixture
def analytics_env_setup(tmp_path):

    analytics_module_path = str(
        pathlib.Path(__file__).resolve().parent / "python_analytics_mockup.py"
    )

    profiles = f"""profiles:
    enrich_one_variable:
        module: {analytics_module_path}
        func: enrich_one_variable
    html_visualization:
        module: {analytics_module_path}
        func: html_visualization
    enrich_multiple_variables:
        module: {analytics_module_path}
        func: enrich_multiple_variables
    enrich_variable_with_arguments:
        module: {analytics_module_path}
        func: enrich_variable_with_arguments
    """

    profile_file = tmp_path / "pythonanalytics.yaml"
    with open(profile_file, "w") as pf:
        pf.write(profiles)

    os.environ["KESTREL_PYTHON_ANALYTICS_CONFIG"] = str(
        profile_file.expanduser().resolve()
    )


@pytest.fixture
def a_session(datasource_env_setup, analytics_env_setup):
    cwd = os.path.dirname(os.path.abspath(__file__))
    bundle = os.path.join(cwd, "test_bundle.json")
    session = Session(debug_mode=True)
    stmt = (
        "conns = get network-traffic"
        f" from file://{bundle}"
        " where [network-traffic:dst_port < 10000]"
    )
    session.execute(stmt)
    return session


@pytest.mark.parametrize(
    "code, expected",
    [
        ("x", []), # No suggestion
        ("c", []), # No suggestion on existing variable name if first token in statement
        ("x ", {"="}),
        (
            "procs = ",
            {"GET", "FIND", "JOIN", "SORT", "GROUP", "LOAD", "NEW", "conns", "_"}
            | TRANSFORMS,
        ),
    ],
)
def test_do_complete_variable_as_first_token(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("disp ", {"conns", "_"} | TRANSFORMS),
        ("disp c", {"onns"}),
        ("disp conns", []),
        ("disp conns ", EXPRESSION_OPTIONS),
        ("disp conns LIMIT 5 ", EXPRESSION_OPTIONS - {"LIMIT"}),
        ("disp conns ATTR name ", EXPRESSION_OPTIONS - {"ATTR"}),
        ("disp conns WHERE name = 'abc' ", EXPRESSION_OPTIONS - {"WHERE"}),
    ],
)
def test_do_complete_disp(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("procs = G", {"ET", "ROUP"}),
        ("urls = g", ["et", "roup"]),
        ("urls = ge", ["t"]),
        ("urls = get ", KNOWN_ETYPES),
        ("urls = get url ", ["FROM", "WHERE"]),
        (
            "urls = GET url FROM ",
            ["_", "conns", "file://", "http://", "https://", "stixshifter://"],
        ),
        ( "urls = GET url FROM stixshi", {"fter://"}),
        ( "urls = GET url FROM stixshifter://", {"thost101", "thost102", "thost103"}),
        ( "urls = GET url FROM stixshifter://thost", {"101", "102", "103"}),
        ("urls = get url where ", []), # TODO: attribute completion
        ("urls = get url where name = 'a' ", {"START"}),
        ("urls = get url where name = 'a' START 2022-01-01T00:00:00Z ", {"STOP"}),
    ],
)
def test_do_complete_cmd_get(datasource_env_setup, a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("procs = F", {"IND"}),
        ("procs = FI", {"ND"}),
        ("procs = FIN", {"D"}),
        ("procs = FIND", []),
        ("procs = FIND ", KNOWN_ETYPES),
        ("procs = FIND p", ["rocess"]),
        ("procs = FIND process", []),
        ("procs = FIND process ", all_relations),
        ("procs = FIND process l", {"oaded", "inked"}),
        ("procs = FIND process c", {"reated", "ontained"}),
        ("procs = FIND process created ", {"conns", "_", "BY"}),
        ("procs = FIND process created BY ", {"conns", "_"}),
        ("procs = FIND process created BY conns ", {"WHERE", "START"}),
    ],
)
def test_do_complete_cmd_find(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("procs2 = SORT procs ", {"BY"}),
        ("procs2 = SORT procs BY ", []), # TODO: attribute completion
    ],
)
def test_do_complete_cmd_sort(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("APPLY ", {"python://", "docker://"}),
        ("APPLY pyth", {"on://"}),
        ("APPLY python://", {"enrich_one_variable", "html_visualization", "enrich_multiple_variables", "enrich_variable_with_arguments"}),
        ("APPLY python://enrich", {"_one_variable", "_multiple_variables", "_variable_with_arguments"}),
        ("APPLY abc ON ", {"_", "conns"}),
        ("APPLY abc ON c ", {"WITH"}),
    ],
)
def test_do_complete_cmd_apply(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("grps = GR", {"OUP"}),
        ("grps = GROUP ", {"conns", "_"}),
        ("grps = GROUP conns ", {"BY"}),
        ("grps = GROUP conns by ", {"BIN"}), # TODO: attribute completion
    ],
)
def test_do_complete_cmd_group(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "code, expected",
    [
        ("procs2 = JOIN x, ", {"_", "conns"}),
        ("procs2 = JOIN x, y ", {"BY"}),
        ("procs2 = JOIN x, y BY ", []), # TODO: attribute completion
        ("procs2 = JOIN x, y BY a", []),
        ("procs2 = JOIN x, y BY a ", {","}),
        ("procs2 = JOIN x, y BY a,", []), # TODO: attribute completion
        ("procs2 = JOIN x, y BY a, ", []), # TODO: attribute completion
    ],
)
def test_do_complete_cmd_join(a_session, code, expected):
    result = a_session.do_complete(code, len(code))
    assert set(result) == set(expected)


@pytest.mark.parametrize(
    "time_string, suffix_ts",
    [
        ("START t'2021", ["-01-01T00:00:00.000Z'"]),
        ("START t'2021-05", ["-01T00:00:00.000Z'"]),
        ("START t'2021-05-04", ["T00:00:00.000Z'"]),
        ("START t'2021-05-04T07:", ["00:00.000Z'"]),
        ("START t'2021-05-04T07:30", [":00.000Z'"]),
        ("START t'2021-05-04T07:30:", ["00.000Z'"]),
        ("START t'2021-05-04T07:30:00Z' STOP t'2021", ["-01-01T00:00:00.000Z'"]),
        ("START t'2021-05-04T07:30:00Z' STOP t'2021-05", ["-01T00:00:00.000Z'"]),
        ("START t'2021-05-04T07:30:00Z' STOP t'2021-05-04", ["T00:00:00.000Z'"]),
        ("START t'2021-05-04T07:30:00Z' STOP t'2021-05-04T07:", ["00:00.000Z'"]),
        ("START t'2021-05-04T07:30:00Z' STOP t'2021-05-04T07:30", [":00.000Z'"]),
        ("START t'2021-05-04T07:30:00Z' STOP t'2021-05-04T07:30:", ["00.000Z'"]),
        ("START 2021", ["-01-01T00:00:00.000Z"]),
        ("START 2021-05", ["-01T00:00:00.000Z"]),
        ("START 2021-05-04", ["T00:00:00.000Z"]),
        ("START 2021-05-04T07:", ["00:00.000Z"]),
        ("START 2021-05-04T07:30", [":00.000Z"]),
        ("START 2021-05-04T07:30:", ["00.000Z"]),
        ("START 2021-05-04T07:30:00Z STOP 2021", ["-01-01T00:00:00.000Z"]),
        ("START 2021-05-04T07:30:00Z STOP 2021-05", ["-01T00:00:00.000Z"]),
        ("START 2021-05-04T07:30:00Z STOP 2021-05-04", ["T00:00:00.000Z"]),
        ("START 2021-05-04T07:30:00Z STOP 2021-05-04T07:", ["00:00.000Z"]),
        ("START 2021-05-04T07:30:00Z STOP 2021-05-04T07:30", [":00.000Z"]),
        ("START 2021-05-04T07:30:00Z STOP 2021-05-04T07:30:", ["00.000Z"]),
    ],
)
def test_session_do_complete_timestamp(time_string, suffix_ts):
    with Session(debug_mode=True) as session:
        script = f"x = GET p WHERE n = 'x' {time_string}"
        result = session.do_complete(script, len(script))
        assert result == suffix_ts



import pytest

from kestrel.session import Session
from kestrel.codegen import prefetch

from .utils import set_empty_kestrel_config, set_no_prefetch_kestrel_config


def test_default_prefetch_config(set_empty_kestrel_config):
    with Session() as s:
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "get", "process"
            )
            == True
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "file"
            )
            == True
        )


def test_prefetch_disabled_config(set_no_prefetch_kestrel_config):
    with Session() as s:
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "get", "process"
            )
            == False
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "file"
            )
            == False
        )



import json
import logging
from typing import Optional
from multiprocessing import Process, Queue, current_process
from typeguard import typechecked

from stix_shifter.stix_translation import stix_translation
from stix_shifter_utils.stix_translation.src.utils.transformer_utils import (
    get_module_transformers,
)
import firepit.aio.ingest

from kestrel_datasource_stixshifter.worker import STOP_SIGN
from kestrel_datasource_stixshifter.worker.utils import TranslationResult, WorkerLog


@typechecked
class Translator(Process):
    def __init__(
        self,
        connector_name: str,
        observation_metadata: dict,
        translation_options: dict,
        cache_data_path_prefix: Optional[str],
        is_fast_translation: bool,
        input_queue: Queue,
        output_queue: Queue,
    ):
        super().__init__()

        self.connector_name = connector_name
        self.observation_metadata = observation_metadata
        self.translation_options = translation_options
        self.cache_data_path_prefix = cache_data_path_prefix
        self.is_fast_translation = is_fast_translation
        self.input_queue = input_queue
        self.output_queue = output_queue

    def run(self):
        worker_name = current_process().name
        translation = stix_translation.StixTranslation()

        for input_batch in iter(self.input_queue.get, STOP_SIGN):
            if input_batch.success:
                if self.is_fast_translation:
                    transformers = get_module_transformers(self.connector_name)

                    mapping = translation.translate(
                        self.connector_name,
                        stix_translation.MAPPING,
                        None,
                        None,
                        self.translation_options,
                    )

                    if "error" in mapping:
                        packet = TranslationResult(
                            worker_name,
                            False,
                            None,
                            WorkerLog(
                                logging.ERROR,
                                f"STIX-shifter mapping failed: {mapping['error']}",
                            ),
                        )
                    else:
                        try:
                            dataframe = firepit.aio.ingest.translate(
                                mapping["to_stix_map"],
                                transformers,
                                input_batch.data,
                                self.observation_metadata,
                            )
                        except Exception as e:
                            packet = TranslationResult(
                                worker_name,
                                False,
                                None,
                                WorkerLog(
                                    logging.ERROR,
                                    f"firepit.aio.ingest.translate() failed with msg: {str(e)}",
                                ),
                            )
                        else:
                            packet = TranslationResult(
                                worker_name,
                                True,
                                dataframe,
                                None,
                            )

                            if self.cache_data_path_prefix:
                                debug_df_filepath = self.get_cache_data_path(
                                    input_batch.offset,
                                    "parquet",
                                )
                                try:
                                    dataframe.to_parquet(debug_df_filepath)
                                except Exception as e:
                                    packet_extra = TranslationResult(
                                        worker_name,
                                        False,
                                        None,
                                        WorkerLog(
                                            logging.ERROR,
                                            f"STIX-shifter fast translation parquet write to disk failed: [{type(e).__name__}] {e}",
                                        ),
                                    )
                                    self.output_queue.put(packet_extra)

                else:
                    stixbundle = translation.translate(
                        self.connector_name,
                        "results",
                        self.observation_metadata,
                        input_batch.data,
                        self.translation_options,
                    )

                    if "error" in stixbundle:
                        packet = TranslationResult(
                            worker_name,
                            False,
                            None,
                            WorkerLog(
                                logging.ERROR,
                                f"STIX-shifter translation to STIX failed: {stixbundle['error']}",
                            ),
                        )
                    else:
                        packet = TranslationResult(
                            worker_name,
                            True,
                            stixbundle,
                            None,
                        )

                    if self.cache_data_path_prefix:
                        debug_stixbundle_filepath = self.get_cache_data_path(
                            input_batch.offset,
                            "json",
                        )
                        try:
                            with open(debug_stixbundle_filepath, "w") as bundle_fp:
                                json.dump(stixbundle, bundle_fp, indent=4)
                        except:
                            packet_extra = TranslationResult(
                                worker_name,
                                False,
                                None,
                                WorkerLog(
                                    logging.ERROR,
                                    f"STIX-shifter translation bundle write to disk failed",
                                ),
                            )
                            self.output_queue.put(packet_extra)

            else:  # rely transmission error/info/debug message
                packet = input_batch

            self.output_queue.put(packet)

        self.output_queue.put(STOP_SIGN)

    def get_cache_data_path(self, offset, suffix):
        offset = str(offset).zfill(32)
        return f"{self.cache_data_path_prefix}_{offset}.{suffix}"



import json
import os

import pytest

from kestrel.exceptions import MissingEntityAttribute
from kestrel.session import Session


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "test_bundle.json")


def test_timestamped_disp(fake_bundle_file):
    with Session() as s:
        stmt = f"""
conns = get network-traffic
        from file://{fake_bundle_file}
        where [network-traffic:dst_port = 22]
"""
        s.execute(stmt)
        out = s.execute("DISP conns")
        data = out[0].to_dict()["data"]
        assert len(data) == 29
        assert "first_observed" not in data[0]
        out = s.execute("DISP TIMESTAMPED(conns)")
        data = out[0].to_dict()["data"]
        assert len(data) == 29
        assert "first_observed" in data[0]
        out = s.execute("DISP TIMESTAMPED(conns) LIMIT 5")
        data = out[0].to_dict()["data"]
        assert len(data) == 5
        assert "first_observed" in data[0]
        out = s.execute(
            "DISP TIMESTAMPED(conns) ATTR first_observed, src_ref.value, src_port"
        )
        data = out[0].to_dict()["data"]
        assert len(data) == 29
        assert "first_observed" in data[0]
        assert "src_ref.value" in data[0]
        assert "src_port" in data[0]
        assert "dst_ref.value" not in data[0]
        assert "dst_port" not in data[0]
        out = s.execute(
            "DISP TIMESTAMPED(conns) ATTR first_observed, src_ref.value, src_port LIMIT 5"
        )
        data = out[0].to_dict()["data"]
        assert len(data) == 5
        assert "first_observed" in data[0]
        assert "src_ref.value" in data[0]
        assert "src_port" in data[0]
        assert "dst_ref.value" not in data[0]
        assert "dst_port" not in data[0]


def test_timestamped_copy(fake_bundle_file):
    with Session() as s:
        stmt = f"""
ips = get ipv4-addr
      from file://{fake_bundle_file}
      where [ipv4-addr:value ISSUBSET '10.0.0.0/8']
ts = timestamped(ips)
"""
        s.execute(stmt)
        ips = s.get_variable("ts")
        print(json.dumps(ips, indent=4))
        assert len(ips) == 100


def test_timestamped_grouped_disp(fake_bundle_file):
    with Session() as s:
        stmt = f"""
conns = GET network-traffic
        FROM file://{fake_bundle_file}
	WHERE [network-traffic:dst_ref.value NOT ISSUBSET '192.168.0.0/16']
grp_conns = GROUP conns BY dst_ref.value WITH COUNT(dst_ref) AS count
DISP TIMESTAMPED(grp_conns)
"""
        with pytest.raises(MissingEntityAttribute) as e:
            s.execute(stmt)


def test_timestamped_grouped_assign(fake_bundle_file):
    with Session() as s:
        stmt = f"""
conns = GET network-traffic
        FROM file://{fake_bundle_file}
	WHERE [network-traffic:dst_ref.value NOT ISSUBSET '192.168.0.0/16']
grp_conns = GROUP conns BY dst_ref.value WITH COUNT(dst_ref) AS count
ts_grp_conns = TIMESTAMPED(grp_conns)
"""
        with pytest.raises(MissingEntityAttribute) as e:
            s.execute(stmt)



"""The STIX-shifter data source package provides access to data sources via
`stix-shifter`_.

The STIX-shifter interface connects to multiple data sources. Users need to
provide one *profile* per data source. The profile name (case insensitive) will
be used in the ``FROM`` clause of the Kestrel ``GET`` command, e.g., ``newvar =
GET entity-type FROM stixshifter://profilename WHERE ...``. Kestrel runtime
will load profiles from 3 places (the later will override the former):

#. STIX-shifter interface config file (only when a Kestrel session starts):

    Create the STIX-shifter interface config file (YAML):

    - Default path: ``~/.config/kestrel/stixshifter.yaml``.

    - A customized path specified in the environment variable ``KESTREL_STIXSHIFTER_CONFIG``.

    Example of STIX-shifter interface config file containing profiles
    (note that the ``options`` section is not required):

    .. code-block:: yaml

        profiles:
            host101:
                connector: elastic_ecs
                connection:
                    host: elastic.securitylog.company.com
                    port: 9200
                    selfSignedCert: false # this means do NOT check cert
                    indices: host101
                    options:  # use any of this section when needed
                        retrieval_batch_size: 10000  # set to 10000 to match default Elasticsearch page size; Kestrel default across connectors: 2000
                        single_batch_timeout: 120  # increase it if hit 60 seconds (Kestrel default) timeout error for each batch of retrieval
                        cool_down_after_transmission: 2  # seconds to cool down between data source API calls, required by some API such as sentinelone; Kestrel default: 0
                        allow_dev_connector: True  # do not check version of a connector to allow custom/testing connector installed with any version; Kestrel default: False
                        dialects:  # more info: https://github.com/opencybersecurityalliance/stix-shifter/tree/develop/stix_shifter_modules/elastic_ecs#dialects
                          - beats  # need it if the index is created by Filebeat/Winlogbeat/*beat
                config:
                    auth:
                        id: VuaCfGcBCdbkQm-e5aOx
                        api_key: ui2lp2axTNmsyakw9tvNnw
            host102:
                connector: qradar
                connection:
                    host: qradar.securitylog.company.com
                    port: 443
                config:
                    auth:
                        SEC: 123e4567-e89b-12d3-a456-426614174000
            host103:
                connector: cbcloud
                connection:
                    host: cbcloud.securitylog.company.com
                    port: 443
                config:
                    auth:
                        org-key: D5DQRHQP
                        token: HT8EMI32DSIMAQ7DJM
        options:  # this section is not required
            fast_translate:  # use firepit-native translation (Dataframe as vessel) instead of stix-shifter result translation (JSON as vessel) for the following connectors
                - qradar
                - elastic_ecs
            translation_workers_count: 8  # default: 2

    Full specifications for data source profile sections/fields:

    - Connector-specific fields: in `stix-shifter`_, go to ``stix_shifter_modules/connector_name/configuration`` like `elastic_ecs config`_.

    - General fields shared across connectors: in `stix-shifter`_, go to `stix_shifter_modules/lang_en.json`_.

#. environment variables (only when a Kestrel session starts):

    Three environment variables are required for each profile:

    - ``STIXSHIFTER_PROFILENAME_CONNECTOR``: the STIX-shifter connector name,
      e.g., ``elastic_ecs``.

    - ``STIXSHIFTER_PROFILENAME_CONNECTION``: the STIX-shifter `connection
      <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#connection>`_
      object in JSON string.

    - ``STIXSHIFTER_PROFILENAME_CONFIG``: the STIX-shifter `configuration
      <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#configuration>`_
      object in JSON string.

    Example of environment variables for a profile:

    .. code-block:: console

        $ export STIXSHIFTER_HOST101_CONNECTOR=elastic_ecs
        $ export STIXSHIFTER_HOST101_CONNECTION='{"host":"elastic.securitylog.company.com", "port":9200, "indices":"host101"}'
        $ export STIXSHIFTER_HOST101_CONFIG='{"auth":{"id":"VuaCfGcBCdbkQm-e5aOx", "api_key":"ui2lp2axTNmsyakw9tvNnw"}}'

#. any in-session edit through the ``CONFIG`` command.

After added data source profiles into ``stixshifter.yaml``, you can test the data source:

.. code-block:: console

    $ stix-shifter-diag data_source_name

where ``data_source_name`` is any profile named in the ``stixshifter.yaml`` config file, usually used in ``FROM stixshifter://data_source_name`` in the ``GET`` command.

The diagnosis utility will check config, test query translation, try connect to the data source to execute a small and a large query, and retrieve data back. Details of all steps will be printed for diagnosis purpose.

If you launch Kestrel in debug mode, STIX-shifter debug mode is still not
enabled by default. To record debug level logs of STIX-shifter, create
environment variable ``KESTREL_STIXSHIFTER_DEBUG`` with any value.

.. _STIX-shifter: https://github.com/opencybersecurityalliance/stix-shifter
.. _elastic_ecs config: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/stix_shifter_modules/elastic_ecs/configuration/lang_en.json
.. _stix_shifter_modules/lang_en.json: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/stix_shifter_modules/lang_en.json

"""

from kestrel.datasource import AbstractDataSourceInterface
from kestrel_datasource_stixshifter.config import load_profiles
from kestrel_datasource_stixshifter.query import query_datasource


class StixShifterInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        """STIX-shifter data source interface only supports ``stixshifter://`` scheme."""
        return ["stixshifter"]

    @staticmethod
    def list_data_sources(config):
        """Get configured data sources from environment variable profiles."""

        # CONFIG command is not supported
        # profiles will be updated according to YAML file and env var
        config["profiles"] = load_profiles()

        data_sources = list(config["profiles"].keys())
        data_sources.sort()
        return data_sources

    @staticmethod
    def query(uri, pattern, session_id, config, store, limit=None):
        """Query a stixshifter data source."""

        return query_datasource(uri, pattern, session_id, config, store, limit)



import pytest

from kestrel.session import Session


def test_copy_and_merge():
    with Session() as s:
        stmt = """
newips = NEW ipv4-addr ["127.0.0.1", "127.0.1.15"]
ip2 = NEW ipv4-addr ["10.0.1.1", "10.0.2.2", "10.0.3.3"]
ip1 = newips
ip3 = ip1 + ip2
"""
        s.execute(stmt)
        newips = s.get_variable("newips")
        assert len(newips) == 2
        assert newips[0]["type"] == "ipv4-addr"
        values = [newips[i]["value"] for i in [0, 1]]
        values.sort()
        assert values == ["127.0.0.1", "127.0.1.15"]

        ip2 = s.get_variable("ip2")
        assert len(ip2) == 3
        assert ip2[0]["type"] == "ipv4-addr"
        values = [row["value"] for row in ip2]
        values.sort()
        assert values == ["10.0.1.1", "10.0.2.2", "10.0.3.3"]

        ip1 = s.get_variable("ip1")
        assert len(ip1) == 2
        assert ip1[0]["type"] == "ipv4-addr"
        values = [row["value"] for row in ip1]
        values.sort()
        assert values == ["127.0.0.1", "127.0.1.15"]

        ip3 = s.get_variable("ip3")
        assert len(ip3) == 5
        assert ip3[0]["type"] == "ipv4-addr"
        values = [row["value"] for row in ip3]
        values.sort()
        assert values == ["10.0.1.1", "10.0.2.2", "10.0.3.3", "127.0.0.1", "127.0.1.15"]


def test_merge_empty_vars():
    with Session() as s:
        stmt = """
newips = NEW ipv4-addr ["127.0.0.1", "127.0.1.15"]
ip1 = GET ipv4-addr FROM newips WHERE value = '10.0.0.1'
ip2 = GET ipv4-addr FROM newips WHERE value = '10.0.0.2'
ip3 = ip1 + ip2
"""
        s.execute(stmt)

        ip3 = s.symtable["ip3"]
        assert len(ip3) == 0
        assert ip3.records_count == 0



import logging
from typing import Optional
from typeguard import typechecked
from contextlib import contextmanager
from multiprocessing import Queue

from kestrel.exceptions import DataSourceError, DataSourceManagerInternalError
from kestrel_datasource_stixshifter.worker.transmitter import TransmitterPool
from kestrel_datasource_stixshifter.worker.translator import Translator
from kestrel_datasource_stixshifter.worker import STOP_SIGN


_logger = logging.getLogger(__name__)


@contextmanager
@typechecked
def transmit(
    connector_name: str,
    connection_dict: dict,
    configuration_dict: dict,
    retrieval_batch_size: int,
    translators_count: int,
    cool_down_after_transmission: int,
    queries: list,
    raw_records_queue: Queue,
    limit: Optional[int],
):
    _logger.debug(f"{translators_count} translation workers to be started")
    transmitter_pool = TransmitterPool(
        connector_name,
        connection_dict,
        configuration_dict,
        retrieval_batch_size,
        translators_count,
        cool_down_after_transmission,
        queries,
        raw_records_queue,
        limit,
    )

    try:
        transmitter_pool.start()
        yield transmitter_pool
    finally:
        transmitter_pool.join(2)
        if transmitter_pool.is_alive():
            raise DataSourceManagerInternalError(
                f"transmitter pool process do not terminate in interface {__package__}"
            )


@contextmanager
@typechecked
def translate(
    connector_name: str,
    observation_metadata: dict,
    translation_options: dict,
    cache_data_path_prefix: Optional[str],
    is_fast_translation: bool,
    raw_records_queue: Queue,
    translated_data_queue: Queue,
    translators_count: int,
):
    _logger.debug(f"fast translation enabled: {is_fast_translation}")
    translators = [
        Translator(
            connector_name,
            observation_metadata,
            translation_options,
            cache_data_path_prefix,
            is_fast_translation,
            raw_records_queue,
            translated_data_queue,
        )
        for _ in range(translators_count)
    ]

    try:
        for translator in translators:
            translator.start()
        yield translators
    finally:
        for translator in translators:
            translator.join(1)
            if translator.is_alive():
                raise DataSourceManagerInternalError(
                    f"one or more translators do not terminate in interface {__package__}"
                )


@typechecked
def read_translated_results(queue: Queue, translator_cnt: int):
    for _ in range(translator_cnt):
        for packet in iter(queue.get, STOP_SIGN):
            if packet.success:
                yield packet.data
            else:
                log_msg = f"[worker: {packet.worker}] {packet.log.log}"
                if packet.log.level == logging.ERROR:
                    _logger.debug(log_msg)
                    raise DataSourceError(log_msg)
                else:
                    if packet.log.level == logging.WARN:
                        _logger.warn(log_msg)
                    elif packet.log.level == logging.INFO:
                        _logger.info(log_msg)
                    else:  # all others as debug logs
                        _logger.debug(log_msg)



import json
from copy import deepcopy
from multiprocessing import Queue
from kestrel.utils import mask_value_in_nested_dict
from kestrel_datasource_stixshifter.config import (
    set_stixshifter_logging_level,
    get_datasource_from_profiles,
    load_options,
    load_profiles,
)
from kestrel_datasource_stixshifter.worker import STOP_SIGN
from kestrel_datasource_stixshifter.query import translate_query
from kestrel_datasource_stixshifter.worker.transmitter import Transmitter
from stix_shifter.stix_transmission import stix_transmission


class Diagnosis:
    def __init__(self, datasource_name):
        self.datasource_name = datasource_name
        self.profiles = load_profiles()
        self.kestrel_options = load_options()
        (
            self.connector_name,
            self.connection_dict,
            self.configuration_dict,
            self.retrieval_batch_size,
            self.cool_down_after_transmission,
            self.allow_dev_connector,
        ) = get_datasource_from_profiles(datasource_name, self.profiles)
        self.if_fast_translation = (
            self.connector_name in self.kestrel_options["fast_translate"]
        )
        set_stixshifter_logging_level()

    def diagnose_config(self):
        print()
        print()
        print()
        print("## Diagnose: config verification")

        configuration_dict_masked = mask_value_in_nested_dict(
            deepcopy(self.configuration_dict)
        )

        print()
        print("#### Kestrel specific config")
        print(f"retrieval batch size: {self.retrieval_batch_size}")
        print(f"cool down after transmission: {self.cool_down_after_transmission}")
        print(f"enable fast translation: {self.if_fast_translation}")

        print()
        print("#### Config to be passed to stix-shifter")
        print(f"connector name: {self.connector_name}")
        print(
            "connection object [ref: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/OVERVIEW.md#connection]:"
        )
        print(json.dumps(self.connection_dict, indent=4))
        print(
            "configuration object [ref: https://github.com/opencybersecurityalliance/stix-shifter/blob/develop/OVERVIEW.md#configuration]:"
        )
        print(json.dumps(configuration_dict_masked, indent=4))

    def diagnose_ping(self):
        print()
        print()
        print()
        print("## Diagnose: stix-shifter to data source connection (network, auth)")

        transmission = stix_transmission.StixTransmission(
            self.connector_name,
            self.connection_dict,
            self.configuration_dict,
        )

        result = transmission.ping()

        print()
        print("#### Results from stixshifter transmission.ping()")
        print(json.dumps(result, indent=4))

    def diagnose_translate_query(self, stix_pattern, quiet=False):
        if not quiet:
            print()
            print()
            print()
            print("## Diagnose: stix-shifter query translation")

        if not quiet:
            print()
            print("#### Input: STIX pattern")
            print(stix_pattern)

        dsl = translate_query(
            self.connector_name,
            {},
            stix_pattern,
            self.connection_dict,
        )

        if "queries" not in dsl:
            raise Exception(str(dsl))

        if not quiet:
            print()
            print(f"#### Output: {len(dsl['queries'])} data source native queries")
            for query in dsl["queries"]:
                print(query)

        return dsl

    def diagnose_run_query_and_retrieval_result(self, stix_patterns, max_batch_cnt):
        print()
        print()
        print()
        print(f"## Diagnose: stix-shifter query execution: <={max_batch_cnt} batch(s)")

        result_queue = Queue()
        result_counts = []

        for pattern in stix_patterns:
            for query in self.diagnose_translate_query(pattern, True)["queries"]:
                transmitter = Transmitter(
                    self.connector_name,
                    self.connection_dict,
                    self.configuration_dict,
                    self.retrieval_batch_size,
                    self.cool_down_after_transmission,
                    query,
                    result_queue,
                    max_batch_cnt * self.retrieval_batch_size,
                )

                transmitter.run()
                result_queue.put(STOP_SIGN)

                print()
                print("#### data retrieval results:")
                for packet in iter(result_queue.get, STOP_SIGN):
                    if packet.success:
                        cnt = len(packet.data)
                        result_counts.append(cnt)
                        print(f"one batch retrieved: {cnt} entries")
                    else:
                        print(packet.log)

                if result_counts:
                    break
                else:
                    print(f"no result matched for query: {query}, go next query")

            if result_counts:
                break
            else:
                print(f"no result matched for pattern: {pattern}, go next pattern")

        return result_counts



import pytest
import os
import sys
import subprocess
import importlib
from importlib.metadata import version

from kestrel.session import Session

from kestrel_datasource_stixshifter.connector import (
    verify_package_origin,
    setup_connector_module,
    get_package_name,
)

from kestrel_datasource_stixshifter.config import get_datasource_from_profiles


def test_verify_package_origin():
    connectors = ["stix_bundle", "qradar", "elastic_ecs", "splunk"]
    for connector_name in connectors:
        verify_package_origin(connector_name, "test_version")


def test_setup_connector_module():
    connectors = ["stix_bundle"]
    for connector_name in connectors:
        setup_connector_module(connector_name)
        importlib.import_module("stix_shifter_modules." + connector_name + ".entry_point")


def test_setup_connector_module_w_wrong_version():
    subprocess.check_call([sys.executable, "-m", "pip", "install", "stix-shifter-modules-paloalto==5.0.0"])
    connector_name = "paloalto"
    setup_connector_module(connector_name)
    importlib.import_module("stix_shifter_modules." + connector_name + ".entry_point")
    stixshifter_version = version("stix_shifter")
    package_name = get_package_name(connector_name)
    package_version = version(package_name)
    assert stixshifter_version == package_version


def test_setup_connector_module_dev_connector():
    subprocess.check_call([sys.executable, "-m", "pip", "install", "stix-shifter-modules-datadog==5.0.0"])
    connector_name = "datadog"
    setup_connector_module(connector_name, True)
    importlib.import_module("stix_shifter_modules." + connector_name + ".entry_point")
    package_version = version(get_package_name(connector_name))
    assert package_version == "5.0.0"


def test_yaml_profiles_refresh(tmp_path):

    profileA = f"""
profiles:
    host101:
        connector: elastic_ecs
        connection:
            host: elastic.securitylog.company.com
            port: 9200
            indices: host101
        config:
            auth:
                id: profileA
                api_key: qwer
"""

    profileB = f"""
profiles:
    host101:
        connector: elastic_ecs
        connection:
            host: elastic.securitylog.company.com
            port: 9200
            indices: host101
            options:
                retrieval_batch_size: 10000
                single_batch_timeout: 120
                cool_down_after_transmission: 5
                allow_dev_connector: True
                dialects:
                    - beats
        config:
            auth:
                id: profileB
                api_key: xxxxxx
"""

    profile_file = tmp_path / "stixshifter.yaml"

    os.environ["KESTREL_STIXSHIFTER_CONFIG"] = str(profile_file.expanduser().resolve())

    with Session() as s:

        with open(profile_file, "w") as pf:
            pf.write(profileA)

        stmt = """
newvar = NEW [ {"type": "process", "name": "cmd.exe", "pid": "123"}
             , {"type": "process", "name": "explorer.exe", "pid": "99"}
             ]
"""
        s.execute(stmt)

        s.data_source_manager.list_data_sources_from_scheme("stixshifter")

        ss_config = s.config["datasources"]["kestrel_datasource_stixshifter"]
        ss_profiles = ss_config["profiles"]
        connector_name, connection, configuration, retrieval_batch_size, cool_down_after_transmission, allow_dev_connector = get_datasource_from_profiles("host101", ss_profiles)
        assert connector_name == "elastic_ecs"
        assert configuration["auth"]["id"] == "profileA"
        assert configuration["auth"]["api_key"] == "qwer"
        assert connection["options"]["timeout"] == 60
        assert connection["options"]["result_limit"] == 2000 * 2
        assert retrieval_batch_size == 2000
        assert cool_down_after_transmission == 0

        with open(profile_file, "w") as pf:
            pf.write(profileB)

        s.data_source_manager.list_data_sources_from_scheme("stixshifter")

        # need to refresh the pointers since the dict is updated
        ss_profiles = ss_config["profiles"]
        connector_name, connection, configuration, retrieval_batch_size, cool_down_after_transmission, allow_dev_connector = get_datasource_from_profiles("host101", ss_profiles)
        assert connector_name == "elastic_ecs"
        assert configuration["auth"]["id"] == "profileB"
        assert configuration["auth"]["api_key"] == "xxxxxx"
        assert connection["options"]["timeout"] == 120
        assert connection["options"]["result_limit"] == 10000 * 2
        assert retrieval_batch_size == 10000
        assert cool_down_after_transmission == 5
        assert allow_dev_connector == True

    del os.environ["KESTREL_STIXSHIFTER_CONFIG"]



import os

from kestrel.session import Session


def test_join_csv_data():
    data_file_path = os.path.join(
        os.path.dirname(__file__), "test_input_data_ips.csv"
    )
    with Session() as s:
        s.execute(f"assets = LOAD {data_file_path} AS ipv4-addr")
        s.execute("""
ips = NEW [{"type": "ipv4-addr", "value": "192.168.1.2"},
           {"type": "ipv4-addr", "value": "192.168.1.3"}]
""")
        s.execute("risk_ips = JOIN ips, assets by value, value")
        v = s.get_variable("risk_ips")
        assert len(v) == 1
        assert v[0]["type"] == "ipv4-addr"
        assert v[0]["value"] == "192.168.1.2"
        assert v[0]["risk"] == 2



import json

import pytest

from kestrel.session import Session


NEW_PROCS = """
procs = NEW [ {"type": "process", "name": "cmd.exe", "pid": 123, "x_foo": "bar"}
            , {"type": "process", "name": "explorer.exe", "pid": 99}]
"""


@pytest.mark.parametrize(
    "attrs, unexpected",
    [
        ("pid", {"name"}),
        ("name", {"pid"}),
        ("pid,name", set()),
    ],
)
def test_expr_attr(attrs, unexpected):
    with Session() as s:
        s.execute(NEW_PROCS)
        out = s.execute(f"DISP procs ATTR {attrs}")
        data = out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))
        actual = set(data[0].keys())
        expected = set(attrs.split(","))
        assert expected == actual
        assert len(unexpected & actual) == 0


@pytest.mark.parametrize(
    "prop, direction, expected",
    [
        ("pid", "asc", [99, 123]),
        ("pid", "desc", [123, 99]),
        ("name", "asc", ["cmd.exe", "explorer.exe"]),
        ("name", "desc", ["explorer.exe", "cmd.exe"]),
    ],
)
def test_expr_sort(prop, direction, expected):
    with Session() as s:
        s.execute(NEW_PROCS)
        out = s.execute(f"DISP procs sort by {prop} {direction}")
        data = out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))
        actual = [p[prop] for p in data]
        assert actual == expected


@pytest.mark.parametrize(
    "limit, offset, expected",
    [
        (5, 0, [99, 123]),
        (1, 0, [99]),
        (2, 1, [123]),
        (1, 1, [123]),
    ],
)
def test_expr_limit_offset(limit, offset, expected):
    with Session() as s:
        s.execute(NEW_PROCS)
        out = s.execute(f"DISP procs SORT BY pid ASC LIMIT {limit} OFFSET {offset}")
        data = out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))
        actual = [p["pid"] for p in data]
        assert actual == expected


@pytest.mark.parametrize(
    "col, op, val, expected",
    [
        ("pid", "=", 99, [99]),
        ("pid", "<", 100, [99]),
        ("pid", ">=", 100, [123]),
        ("x_foo", "IS NULL", "", [99]),
        ("x_foo", "IS NOT NULL", "", [123]),
        ("x_foo", "=", "'bar'", [123]),
    ],
)
def test_expr_where(col, op, val, expected):
    with Session() as s:
        s.execute(NEW_PROCS)
        for stmt in (
            f"DISP procs WHERE {col} {op} {val}",
            f"DISP procs WHERE [{col} {op} {val}]",
        ):
            out = s.execute(stmt)
            data = out[0].to_dict()["data"]
            print(json.dumps(data, indent=4))
            actual = [p["pid"] for p in data]
            assert actual == expected


def test_expr_assign_where():
    with Session() as s:
        s.execute(NEW_PROCS)
        out = s.execute("x = procs WHERE pid > 100")
        data = out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))
        vars_updated = data["variables updated"]
        assert vars_updated[0]["#(ENTITIES)"] == 1
